
import "@babylonjs/core/Debug/debugLayer";
import "@babylonjs/inspector";
import "@babylonjs/loaders/glTF";
import "@babylonjs/controls";
import { AssetsManager, Engine, Scene, FreeCamera, Vector3, HemisphericLight, PointLight, SpotLight, DirectionalLight, Mesh, VertexData, StandardMaterial, Texture, SceneLoader, ScreenshotTools, VertexBuffer, Viewport, Color4, ArcRotateCamera } from "@babylonjs/core";
import {GLTF2Export} from "@babylonjs/serializers";
import { useCallback } from "react";
import * as fs from 'fs';
import * as path from 'path';
import Jimp from 'jimp';

const MeshNamePage_1:string = "Page-1";
const MeshNamePage_2:string = "Page-2";
const TextureNamePage_1:string = "TexturePage-1";
const TextureNamePage_2:string = "TexturePage-2";
const PageTextureName:string = "PageTexture";

function exportPageMesh(exportFileName:string, scene:Scene):void{
    let options = {
        shouldExportNode: function (node:any) {
          return node["name"] == MeshNamePage_1 || node["name"] == MeshNamePage_2;
        },
      };
      
      GLTF2Export.GLBAsync(scene, exportFileName, options).then((glb) => {
        glb.downloadFiles();
      });
}

function mapTexture(scene:Scene, pageTexture:Texture[]){
    var positionPage_1 = scene.meshes[1].getVerticesData(VertexBuffer.PositionKind);
    var positionPage_2 = scene.meshes[2].getVerticesData(VertexBuffer.PositionKind);
    if (positionPage_1 == null || positionPage_2 == null){
        return;
    }
    var vertexCount = positionPage_1.length/3;
    var uvs1:any = [];
    var uvs2:any = [];
    for ( var i =0; i< vertexCount; i++) {
        uvs1[2*i] = 0.33*(positionPage_1[3*i] - positionPage_1[0]);
        uvs1[2*i + 1] = (positionPage_1[3*i + 1] - positionPage_1[1]);

        uvs2[2*i] = 0.33*(positionPage_2[3*i] - positionPage_2[0]);
        uvs2[2*i + 1] = (positionPage_2[3*i + 1] - positionPage_2[1]);
    }
    scene.meshes[1].setVerticesData(VertexBuffer.UVKind, uvs1);
    scene.meshes[2].setVerticesData(VertexBuffer.UVKind, uvs2);

    var material_1 = new StandardMaterial(TextureNamePage_1, scene);
    material_1.diffuseTexture = pageTexture[0];
    scene.meshes[1].material = material_1;

    var material_2 = new StandardMaterial(TextureNamePage_2, scene);
    material_2.diffuseTexture = pageTexture[1];
    scene.meshes[2].material = material_2;

}

class App {    

    addBookPageMesh(xSize:number, ySize:number, xStartPosition:number, yStartPosition:number, meshName:string, scene:Scene):Mesh {
        var meshSize = 100;
        var xStep = xSize/meshSize;
        var yStep = ySize/meshSize;
        
        var counter = 0;
        var positions = [];
        for (var i = 0; i< meshSize; i++) {
            for (var j = 0; j< meshSize; j++) {
            
                positions[counter] = xStartPosition + j*xStep;
                positions[counter + 1] = yStartPosition + i*yStep;
                positions[counter + 2] = 0;
                
                positions[counter + 3] = xStartPosition + (j + 1)*xStep ;
                positions[counter + 4] = yStartPosition + i*yStep;
                positions[counter + 5] = 0;

                positions[counter + 6] = xStartPosition + j*xStep;
                positions[counter + 7] = yStartPosition + (i + 1)*yStep;
                positions[counter + 8] = 0;

                positions[counter + 9] = xStartPosition + j*xStep;
                positions[counter + 10] = yStartPosition + (i + 1)*yStep;
                positions[counter + 11] = 0;

                positions[counter + 12] = xStartPosition+ (j + 1)*xStep ;
                positions[counter + 13] = yStartPosition + i*yStep;
                positions[counter + 14] = 0; 

                positions[counter + 15] = xStartPosition+ (j + 1)*xStep;
                positions[counter + 16] =  yStartPosition + (i + 1)*yStep;
                positions[counter + 17] = 0;

                counter += 18;
            }
        }

        var vertex_count  = counter/3;
        var indices = [];
        for ( var i =0; i< vertex_count; i++) {
            indices[i] = i;
        }

        var uvs = [];
        for ( var i =0; i< vertex_count; i++) {
            uvs[2*i] = (positions[3*i] - positions[0])/xSize;
            uvs[2*i + 1] = (positions[3*i + 1] - positions[1])/ySize;
        }

        //Empty array to contain calculated values or normals added
        var normals:any = [];
        VertexData.ComputeNormals(positions, indices, normals);

        //Calculations of normals added
        var vertexData = new VertexData();
        vertexData.positions = positions;
        vertexData.indices = indices;
        vertexData.normals = normals;
        vertexData.uvs = uvs;        
        var bookPageMesh = new Mesh(meshName, scene);
        vertexData.applyToMesh(bookPageMesh, true);

        var material = new StandardMaterial(PageTextureName, scene);
        material.wireframe = true;
        bookPageMesh.material = material;

        return bookPageMesh;
    }

    applyBookFlat_SlantOpen(meshPage_1:Mesh, meshPage_2:Mesh, leftRightWeight:number, applyOnLeft:boolean, applyOnRight:boolean){
        // Update poistion
        var positionPage_1 = meshPage_1.getVerticesData(VertexBuffer.PositionKind);
        var positionPage_2 = meshPage_2.getVerticesData(VertexBuffer.PositionKind);

        if (positionPage_1 == null || positionPage_2 == null){
            return;
        }
        var vertexCount = positionPage_1.length/3;
        for (var i = 0; i< vertexCount; i++){
            if(applyOnLeft){
                positionPage_1[3*i + 2] =  -leftRightWeight*(1- positionPage_1[3*i]);
            }
            if(applyOnRight){
                positionPage_2[3*i  + 2] =   -(1-leftRightWeight)*(positionPage_2[3*i] - 1);
            }
        }

        if(applyOnLeft){
            meshPage_1.updateVerticesData(VertexBuffer.PositionKind, positionPage_1);
            // Update normal
            var normalsPage_1:any = [];
            VertexData.ComputeNormals(positionPage_1, meshPage_1.getIndices(), normalsPage_1);
            meshPage_1.updateVerticesData(VertexBuffer.NormalKind, normalsPage_1);
        }

        if(applyOnRight){
            meshPage_2.updateVerticesData(VertexBuffer.PositionKind, positionPage_2);
            var normalsPage_2:any = [];
            VertexData.ComputeNormals(positionPage_2, meshPage_2.getIndices(), normalsPage_2);
            meshPage_2.updateVerticesData(VertexBuffer.NormalKind, normalsPage_2);
        }
    }



    applyBookCurvePerfectBinding_MidOpen_BothCurve(meshPage_1:Mesh, meshPage_2:Mesh, curveOffset:number, leftRightWeight:number, applyOnLeft:boolean, applyOnRight:boolean){
        // Update poistion
        var positionPage_1 = meshPage_1.getVerticesData(VertexBuffer.PositionKind);
        var positionPage_2 = meshPage_2.getVerticesData(VertexBuffer.PositionKind);

        if (positionPage_1 == null || positionPage_2 == null){
            return;
        }
        var vertexCount = positionPage_1.length/3;
        for (var i = 0; i< vertexCount; i++){
            if(applyOnLeft){
                positionPage_1[3*i + 2] =  2*leftRightWeight*(positionPage_1[3*i] + curveOffset) *(positionPage_1[3*i] - 0.33);
            }

            if(applyOnRight){
                positionPage_2[3*i  + 2] =   2*(1-leftRightWeight)*(positionPage_2[3*i] - 0.33) *(positionPage_2[3*i] - 0.66 - curveOffset);
            }
        }

        if(applyOnLeft){
            meshPage_1.updateVerticesData(VertexBuffer.PositionKind, positionPage_1);
             // Update normal
            var normalsPage_1:any = [];
            VertexData.ComputeNormals(positionPage_1, meshPage_1.getIndices(), normalsPage_1);
            meshPage_1.updateVerticesData(VertexBuffer.NormalKind, normalsPage_1);
        }
        
        if(applyOnRight){
            meshPage_2.updateVerticesData(VertexBuffer.PositionKind, positionPage_2);
            var normalsPage_2:any = [];
            VertexData.ComputeNormals(positionPage_2, meshPage_2.getIndices(), normalsPage_2);
            meshPage_2.updateVerticesData(VertexBuffer.NormalKind, normalsPage_2);
        }
    }
    
    
    applyBookCurvePerfectBinding_MidOpen_BothCurve_PageFlattens(meshPage_1:Mesh, meshPage_2:Mesh, curveOffset:number, leftRightWeight:number, applyOnLeft:boolean, applyOnRight:boolean){
        // Update poistion
        var positionPage_1 = meshPage_1.getVerticesData(VertexBuffer.PositionKind);
        var positionPage_2 = meshPage_2.getVerticesData(VertexBuffer.PositionKind);

        if (positionPage_1 == null || positionPage_2 == null){
            return;
        }
        var vertexCount = positionPage_1.length/3;
        for (var i = 0; i< vertexCount; i++){
            if(applyOnLeft){
                positionPage_1[3*i + 2] =  -1 + 1/(1 + Math.exp(15*(0.75- positionPage_1[3*i])));
            }

            if(applyOnRight){
                positionPage_2[3*i  + 2] =   -1 + 1/(1 + Math.exp(15*( positionPage_2[3*i] -0.75)));
            }
        }

        if(applyOnLeft){
            meshPage_1.updateVerticesData(VertexBuffer.PositionKind, positionPage_1);
             // Update normal
            var normalsPage_1:any = [];
            VertexData.ComputeNormals(positionPage_1, meshPage_1.getIndices(), normalsPage_1);
            meshPage_1.updateVerticesData(VertexBuffer.NormalKind, normalsPage_1);
        }
        
        if(applyOnRight){
            meshPage_2.updateVerticesData(VertexBuffer.PositionKind, positionPage_2);
            var normalsPage_2:any = [];
            VertexData.ComputeNormals(positionPage_2, meshPage_2.getIndices(), normalsPage_2);
            meshPage_2.updateVerticesData(VertexBuffer.NormalKind, normalsPage_2);
        }
    }

    applyBookCurvePerfectBinding_MidOpen_PageCurve_starting(meshPage_1:Mesh, meshPage_2:Mesh, curveOffset:number, leftRightWeight:number, applyOnLeft:boolean, applyOnRight:boolean){
        // Update poistion
        var positionPage_1 = meshPage_1.getVerticesData(VertexBuffer.PositionKind);
        var positionPage_2 = meshPage_2.getVerticesData(VertexBuffer.PositionKind);

        if (positionPage_1 == null || positionPage_2 == null){
            return;
        }
        var vertexCount = positionPage_1.length/3;
        for (var i = 0; i< vertexCount; i++){
            if(applyOnLeft){
                var fact = 1;  
                for (var k = 1; k<= positionPage_1[i-2]; k++)
                {
                    fact *= k;
                }
                
                positionPage_1[3*i+2]= -20*leftRightWeight*(Math.pow(3, positionPage_1[3*i])*Math.pow(2.71828, -3))/(fact);
        
            }

            if(applyOnRight){
                
                var fact = 1;     
                for (var k = 1; k<= positionPage_2[i-2]; k++)
                {
                    fact *= k;
                }
        
                positionPage_2[3*i+2]= -20*(1-leftRightWeight)*(Math.pow(3, (positionPage_2[3*i-1]))*Math.pow(2.71828, -3))/(fact);    
                
            }
        }

        if(applyOnLeft){
            meshPage_1.updateVerticesData(VertexBuffer.PositionKind, positionPage_1);
             // Update normal
            var normalsPage_1:any = [];
            VertexData.ComputeNormals(positionPage_1, meshPage_1.getIndices(), normalsPage_1);
            meshPage_1.updateVerticesData(VertexBuffer.NormalKind, normalsPage_1);
        }
        
        if(applyOnRight){
            meshPage_2.updateVerticesData(VertexBuffer.PositionKind, positionPage_2);
            var normalsPage_2:any = [];
            VertexData.ComputeNormals(positionPage_2, meshPage_2.getIndices(), normalsPage_2);
            meshPage_2.updateVerticesData(VertexBuffer.NormalKind, normalsPage_2);
        }
    }

    createBookObject(scene:Scene, assetsManager:AssetsManager){
        var bookPage_1: Mesh = this.addBookPageMesh(0.33, 1, 0, 0, MeshNamePage_1, scene);
        var bookPage_2: Mesh = this.addBookPageMesh(0.33, 1, 0.33, 0, MeshNamePage_2, scene);
        //this.applyBookCurvePerfectBinding_MidOpen_BothCurve_PageFlattens(bookPage_1, bookPage_2, 0.5, 0.5, true, true);
        this.applyBookCurvePerfectBinding_MidOpen_BothCurve(bookPage_1, bookPage_2, 0.5, 0.5, true, true);
        //this.applyBookCurvePerfectBinding_MidOpen_PageCurve_starting(bookPage_1, bookPage_2, 0.5, 0.5, true, true);
    }

    addLight(scene:Scene, lightType:number){
        if(lightType == 1){
            var pointLight = new PointLight("pointLight", new Vector3(5, 5, -5), scene);
        }
        if(lightType == 2){
            var directionalLight = new DirectionalLight("directionalLight", new Vector3(1, 1, 1), scene);
        }
        if(lightType == 3){
            var spotLight = new SpotLight("spotLight", new Vector3(0, 0, -5), new Vector3(0, 0, 1), Math.PI / 3, 2, scene);
        }
        if(lightType == 4){
            var ambinetLight = new HemisphericLight("HemiLight", new Vector3(1, 0.5, 0), scene);
        }
        if(lightType == 5){
            var ambinetLight = new HemisphericLight("HemiLight", new Vector3(1, 0.5, 0), scene);
            var pointLight = new PointLight("pointLight", new Vector3(5, 5, -5), scene);
        }
        if(lightType == 6){
            var ambinetLight = new HemisphericLight("HemiLight", new Vector3(1, 0.5, 0), scene);
            var spotLight = new SpotLight("spotLight", new Vector3(0, 0, -10), new Vector3(0, 0, 1), Math.PI / 3, 2, scene);
        }
        if(lightType == 7){
            var pointLight_1 = new PointLight("pointLight_1", new Vector3(5, 5, -10), scene);
            var pointLight_2 = new PointLight("pointLight_2", new Vector3(-5, -5, -10), scene);
        }
    }
    
    constructor() {
        // create the canvas html element and attach it to the webpage
        var canvas = document.createElement("canvas");
        canvas.style.width = "100%";
        canvas.style.height = "100%";
        canvas.id = "gameCanvas";
        document.body.appendChild(canvas);

        // initialize babylon scene and engine
        var engine = new Engine(canvas, true);
        var scene = new Scene(engine);
        //scene.clearColor = new Color4(0, 0, 0, 0);
        let assetsManager = new AssetsManager(scene);

        // This creates and positions a free camera (non-mesh)
        //var cameraPosition = new Vector3(1.0, 0.5, -5);
        var cameraPosition = new Vector3(1.0, 0.5, -2.2);
        var camera = new FreeCamera("camera1", cameraPosition, scene);
        
        camera.setTarget(new Vector3(1.0, 0.5, 0));
        camera.maxZ = 12;
        //camera.minZ = 1;
        
        camera.attachControl(canvas, true);
        this.addLight(scene, 1);
        this.createBookObject(scene, assetsManager);
        var pageTexture:any = [];

        var meshTask = assetsManager.addMeshTask("meshTask", [MeshNamePage_1, MeshNamePage_2], "assets/AR-3-4/", "Book_Perfect_Binding_MidOpen_BothCurve_ThicknessOffset.glb");
        //var meshTask = assetsManager.addMeshTask("meshTask", [MeshNamePage_1, MeshNamePage_2], "assets/AR-1-3/", "Book_midOpen_pageCurve.glb");
        meshTask.onSuccess = function(task){
            try{
                mapTexture(scene, pageTexture);
                    
            }
            catch(error) {
                console.log("Texture mapping failed" + error);
            }           
        }
        meshTask.onError= function(task){
                console.log("Mesh task failed");
        }

        var textureTaskPage_1 = assetsManager.addTextureTask("textureTaskPage_1", "assets/Page-3.png");
        textureTaskPage_1.onSuccess = function(task) {
            pageTexture[0] = task.texture;
        };
        textureTaskPage_1.onError = function(task2) {
            console.log("Texture load failed");
        };

        var textureTaskPage_2 = assetsManager.addTextureTask("textureTaskPage_2", "assets/Page-3.png");
        textureTaskPage_2.onSuccess = function(task) {
            pageTexture[1] = task.texture;            
        };       

        window.addEventListener("keydown", (ev) => {
            // Shift+Ctrl+Alt+I
            if (ev.shiftKey && ev.ctrlKey && ev.altKey && ev.keyCode === 73) {
                if (scene.debugLayer.isVisible()) {
                    scene.debugLayer.hide();
                } else {
                    scene.debugLayer.show();
                }
            }
        });

        window.setTimeout(function () { 
            var currentCanvasSizeWidth = canvas.style.width;
            var currentCanvasSizeHeight = canvas.style.height;
            canvas.style.width = "1920px";
            canvas.style.height = "1080px";
            
            engine.resize();
            scene.render();        
            ScreenshotTools.CreateScreenshotAsync(engine, camera, {width:1920, height:1080}).then((data) => {
                const a = document.createElement('a');
                //a.setAttribute('download', 'snapshot.png');
                a.setAttribute('href', data);
                a.click();
                
                canvas.style.width = currentCanvasSizeWidth;
                canvas.style.height = currentCanvasSizeHeight;
                engine.resize();

            });
        }, 5000);

        //this.crop_image('snapshot.png');
       
        assetsManager.onFinish = function (tasks) {
            //exportPageMesh("Book_midOpen_pageCurve.glb", scene);
            engine.runRenderLoop(function () {
                scene.render();
            });
        };

        assetsManager.load();
    }

    // async crop_image(image_name:string){

    //     const dir_path = "C:\\Users\\t-jreddivari\\Desktop\\Microsoft\\MeshGeneration\\MeshGeneration\\Samples\\Lights\\";
    //     fs.readdir(dir_path, function(err, files){
    //         if (err){
    //             return console.log('Unable to scan directory: '+ err);
    //         }
    //         const image = await Jimp.read(path.join(dir_path, image_name));
    //         image.crop(25, 695, 1050, 536, function(err){
    //             if (err) throw err;
    //           })
    //           .write('crop1.png');
    // }
}

new App();
